from .actions import (  # noqa: F401
    broadcast,
    checkpoint,
    connect_to_endpoints,
    request,
    serve_request,
    throws,
    wait_any_then_broadcast,
    wait_for,
    wait_until_any_endpoint_subscribed_to,
    wait_until_connected_to,
)
from .driver import driver  # noqa: F401
from .initializers import run_endpoint, serve_endpoint  # noqa: F401



from .asyncio import AsyncioEndpoint  # noqa: F401
from .base import EndpointAPI  # noqa: F401
from .common import (  # noqa: F401
    BaseEvent,
    BaseRequestResponseEvent,
    BroadcastConfig,
    ConnectionConfig,
    Subscription,
)
from .exceptions import ConnectionAttemptRejected, UnexpectedResponse  # noqa: F401
from .trio import TrioEndpoint  # noqa: F401



import datetime
from logging import Logger
import time

from ..backends import BaseBackend
from ..stats import GlobalStatistic, Total


def print_global_header(logger: Logger, backend: BaseBackend) -> None:

    logger.info(f'{"+++" + backend.name.upper() + "+++":^150}')
    logger.info(f'{"+++Globals+++":^150}')
    logger.info(
        "|{:^19}|{:^16}|{:^23}|{:^20}|{:^16}|{:^16}|{:^16}|".format(
            "Consumer processes",
            "Total time",
            "Total aggegated time",
            "Propagated events",
            "Received events",
            "Propagated EPS",
            "Received EPS",
        )
    )


def print_global_entry(
    logger: Logger,
    num_consumer_processes: int,
    num_events: int,
    global_statistic: GlobalStatistic,
) -> None:

    total_propagates_events = num_events
    total_duration = time.time() - global_statistic._created_at
    total_aggregated_time = sum(
        n.total_aggregated_time for n in global_statistic._entries
    )
    total_received_events = sum(n.num_total for n in global_statistic._entries)
    propagated_eps = total_propagates_events / total_duration
    received_eps = total_received_events / total_duration

    logger.info(
        "|{:^19}|{:^16.5f}|{:^23.5f}|{:^20}|{:^16}|{:^16.3f}|{:^16.3f}|".format(
            num_consumer_processes,
            total_duration,
            total_aggregated_time,
            total_propagates_events,
            total_received_events,
            propagated_eps,
            received_eps,
        )
    )


def print_entry_header(logger: Logger) -> None:
    logger.info(f'{"+++Process Details+++":^150}')
    logger.info(
        "|{:^19}|{:^16}|{:^16}|{:^16}|{:^16}|{:^16}|{:^16}|{:^16}|{:^23}|".format(
            "Process",
            "Processed Events",
            "First sent",
            "Last received",
            "Fastest",
            "Slowest",
            "AVG",
            "Total duration",
            "Total aggregated time",
        )
    )


def to_readable_timestamp(timestamp: float) -> str:
    return datetime.datetime.utcfromtimestamp(timestamp).strftime("%H:%M:%S.%f")[:-3]


def print_entry_line(logger: Logger, total: Total) -> None:
    logger.info(
        "|{:^19}|{:^16}|{:^16}|{:^16}|{:^16.5f}|{:^16.5f}|{:^16.5f}|{:^16.5f}|{:^23.5f}|".format(  # noqa: E501
            total.caption,
            total.num_total,
            to_readable_timestamp(total.first_sent),
            to_readable_timestamp(total.last_received),
            total.duration_fastest,
            total.duration_slowest,
            total.duration_avg,
            total.total_duration,
            total.total_aggregated_time,
        )
    )


def print_full_report(
    logger: Logger,
    backend: BaseBackend,
    num_consumer_processes: int,
    num_events: int,
    global_statistic: GlobalStatistic,
) -> None:

    print_global_header(logger, backend)
    print_global_entry(logger, num_consumer_processes, num_events, global_statistic)

    print_entry_header(logger)
    for total in global_statistic._entries:
        print_entry_line(logger, total)



import asyncio
import functools
import logging
import multiprocessing
from typing import Any, Awaitable, Callable, Optional, Sequence, Tuple, Type, Union

from lahja import (
    BaseEvent,
    BaseRequestResponseEvent,
    BroadcastConfig,
    ConnectionConfig,
    EndpointAPI,
)
from lahja.base import EventAPI
from lahja.exceptions import LahjaError
from lahja.tools.engine import EngineAPI

logger = logging.getLogger("lahja.driver")


class DidNotThrow(LahjaError):
    pass


class SyncAction:
    def __str__(self) -> str:
        return self._fn.__name__

    def __init__(self, fn: Callable[..., None], **kwargs: Any) -> None:
        functools.update_wrapper(self, fn)
        self._fn = fn
        self._kwargs = kwargs

    def __call__(self, engine: EngineAPI, endpoint: EndpointAPI) -> None:
        self._fn(engine, endpoint, **self._kwargs)


class AsyncAction:
    def __str__(self) -> str:
        return self._coro.__name__

    def __init__(self, coro: Callable[..., Awaitable[None]], **kwargs: Any) -> None:
        functools.update_wrapper(self, coro)
        self._coro = coro
        self._kwargs = kwargs

    async def __call__(self, engine: EngineAPI, endpoint: EndpointAPI) -> None:
        await self._coro(engine, endpoint, **self._kwargs)


Action = Union[SyncAction, AsyncAction]
EventHandlerFn = Callable[[EndpointAPI, BaseEvent], Any]


#
# EndpointAPI.wait_for
#
async def _wait_for(
    engine: EngineAPI,
    endpoint: EndpointAPI,
    event_type: Type[BaseEvent],
    on_event: Optional[EventHandlerFn],
) -> None:
    logger.debug("[%s].wait_for(%s)", endpoint, event_type)
    result = await endpoint.wait_for(event_type)
    logger.debug("[%s].wait_for(%s) RECEIVED", endpoint, event_type)
    if on_event is not None:
        if asyncio.iscoroutinefunction(on_event):
            await on_event(endpoint, result)
        else:
            on_event(endpoint, result)


def wait_for(
    event_type: Type[BaseEvent], on_event: Optional[EventHandlerFn] = None
) -> AsyncAction:
    """
    Wait for an event of the provided ``request_type`` and call
    response event returned by the provide ``get_response`` function.
    """
    return AsyncAction(_wait_for, event_type=event_type, on_event=on_event)


#
# Serve request
#
GetResponseFn = Callable[[EndpointAPI, BaseRequestResponseEvent[BaseEvent]], BaseEvent]


def serve_request(
    request_type: Type[BaseRequestResponseEvent[BaseEvent]], get_response: GetResponseFn
) -> AsyncAction:
    """
    Wait for an event of the provided ``request_type`` and respond using the
    response event returned by the provide ``get_response`` function.
    """

    async def _serve_response(
        endpoint: EndpointAPI, request: BaseRequestResponseEvent[BaseEvent]
    ) -> None:
        response = get_response(endpoint, request)
        await endpoint.broadcast(response, config=request.broadcast_config())
        logger.debug("[%s] sent response: %s", endpoint, response)

    return AsyncAction(_wait_for, event_type=request_type, on_event=_serve_response)


#
# EndpointAPI.request
#
async def _request(
    engine: EngineAPI,
    endpoint: EndpointAPI,
    event: BaseRequestResponseEvent[BaseEvent],
    config: Optional[BroadcastConfig],
    on_response: Optional[EventHandlerFn],
) -> None:
    logger.debug("[%s].request(%s, config=%s)", endpoint, event, config)
    response = await endpoint.request(event, config=config)
    logger.debug("[%s].request(%s) GOT RESPONSE: %s", endpoint, event, response)
    if on_response is not None:
        if asyncio.iscoroutinefunction(on_response):
            await on_response(endpoint, response)
        else:
            on_response(endpoint, response)


def request(
    event: BaseRequestResponseEvent[BaseEvent],
    config: Optional[BroadcastConfig] = None,
    on_response: Optional[EventHandlerFn] = None,
) -> AsyncAction:
    """
    see ``EndpointAPI.connect_to_endpoints``

    Optionally provide a callback ``on_response`` that will be run upon receipt
    of the response.
    """
    return AsyncAction(_request, event=event, config=config, on_response=on_response)


#
# EndpointAPI.connect_to_endpoints
#
async def _connect_to_endpoints(
    engine: EngineAPI, endpoint: EndpointAPI, configs: Sequence[ConnectionConfig]
) -> None:
    logger.debug(
        "[%s].connect_to_endpoints(%s)",
        endpoint,
        ",".join((str(config) for config in configs)),
    )
    await endpoint.connect_to_endpoints(*configs)


def connect_to_endpoints(*configs: ConnectionConfig) -> AsyncAction:
    """
    see ``EndpointAPI.connect_to_endpoints``
    """
    return AsyncAction(_connect_to_endpoints, configs=configs)


#
# EndpointAPI.wait_until_any_endpoint_subscribed_to
#
async def _wait_until_any_endpoint_subscribed_to(
    engine: EngineAPI, endpoint: EndpointAPI, event_type: Type[BaseEvent]
) -> None:
    logger.debug("[%s].wait_until_any_endpoint_subscribed_to(%s)", endpoint, event_type)
    await endpoint.wait_until_any_endpoint_subscribed_to(event_type)


def wait_until_any_endpoint_subscribed_to(event_type: Type[BaseEvent]) -> AsyncAction:
    """
    see ``EndpointAPI.wait_until_any_endpoint_subscribed_to``
    """
    return AsyncAction(_wait_until_any_endpoint_subscribed_to, event_type=event_type)


#
# EndpointAPI.wait_until_connected_to
#
async def _wait_until_connected_to(
    engine: EngineAPI, endpoint: EndpointAPI, name: str
) -> None:
    logger.debug("[%s].wait_until_connected_to(%s)", endpoint, name)
    await endpoint.wait_until_connected_to(name)


def wait_until_connected_to(name: str) -> AsyncAction:
    """
    see ``EndpointAPI.wait_until_connected_to``
    """
    return AsyncAction(_wait_until_connected_to, name=name)


#
# EndpointAPI.broadcast
#
async def _broadcast(
    engine: EngineAPI,
    endpoint: EndpointAPI,
    event: BaseEvent,
    config: Optional[BroadcastConfig],
) -> None:
    logger.debug("[%s].broadcast(%s, config=%s)", endpoint, event, config)
    await endpoint.broadcast(event, config=config)


def broadcast(
    event: BaseEvent, config: Optional[BroadcastConfig] = None
) -> AsyncAction:
    """
    see ``EndpointAPI.broadcast``
    """
    return AsyncAction(_broadcast, event=event, config=config)


#
# EndpointAPI wait then broadcast
#
async def _wait_any_then_broadcast(
    engine: EngineAPI,
    endpoint: EndpointAPI,
    event: BaseEvent,
    config: Optional[BroadcastConfig],
) -> None:
    # TODO: these could be abstracted into a single `compose` action which
    # combines multiple actions.
    await _wait_until_any_endpoint_subscribed_to(engine, endpoint, type(event))
    await _broadcast(engine, endpoint, event, config)


def wait_any_then_broadcast(
    event: BaseEvent, config: Optional[BroadcastConfig] = None
) -> AsyncAction:
    """
    Combination of ``wait_until_any_endpoint_subscribed_to`` and ``broadcast``
    """
    return AsyncAction(_wait_any_then_broadcast, event=event, config=config)


#
# Throws
#
async def _throws(
    engine: EngineAPI, endpoint: EndpointAPI, action: Action, exc_type: Type[Exception]
) -> None:
    logger.debug("[%s](%s) - Expecting Error: %s", endpoint, action, exc_type)
    if not isinstance(action, (SyncAction, AsyncAction)):
        # We have to do this up here to ensure we don't end up catching in the
        # try/except below.
        raise TypeError(f"Unsupported action: {action}")

    try:
        if isinstance(action, AsyncAction):
            await action(engine, endpoint)
        elif isinstance(action, SyncAction):
            action(engine, endpoint)
        else:
            raise Exception("Unreachable code path")
    except exc_type:
        logger.debug("[%s](%s) - Got Error: %s", endpoint, action, exc_type)
    else:
        logger.debug("[%s](%s) - Did Not Error: %s", endpoint, action, exc_type)
        raise DidNotThrow(f"Action `{action}` did not throw expected error: {exc_type}")


def throws(action: Action, exc_type: Type[Exception]) -> Action:
    """
    Checks that the provided *Action* throws the provided exception type.
    """
    return AsyncAction(_throws, action=action, exc_type=exc_type)


#
# Synchronization Point
#
async def _checkpoint(
    engine: EngineAPI,
    endpoint: EndpointAPI,
    name: str,
    my_event: EventAPI,
    other_event: EventAPI,
) -> None:
    logger.debug("[%s] at checkpoint: %s", endpoint, name)
    my_event.set()
    while not other_event.is_set():
        await engine.sleep(0.01)


def checkpoint(name: str) -> Tuple[AsyncAction, AsyncAction]:
    """
    Generates a pair of actions that can be used in separate drivers to
    synchronize their action execution.  Each driver will wait until this
    checkpoint has been hit before proceeding.
    """
    left = multiprocessing.Event()
    right = multiprocessing.Event()
    return (
        AsyncAction(_checkpoint, name=name, my_event=left, other_event=right),
        AsyncAction(_checkpoint, name=name, my_event=right, other_event=left),
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import (
    setup,
    find_packages,
)

extras_require = {
    'test': [
        "cytoolz>=0.9.0,<1.0.0",
        "pytest==4.0.2",
        "pytest-timeout==1.3.3",
        "pytest-xdist==1.25.0",
        "tox>=2.9.1,<3",
    ],
    'test-asyncio': [
        "pytest-asyncio==0.9.0",
    ],
    'test-trio': [
        "pytest-trio==0.5.2",
    ],
    'lint': [
        "black==19.3b",
        "flake8==3.7.7",
        "isort==4.3.18",
        "mypy==0.701",
    ],
    'doc': [
        "Sphinx>=1.6.5,<2",
        "sphinx_rtd_theme>=0.1.9",
        "towncrier>=19.2.0, <20",
    ],
    'dev': [
        "bumpversion>=0.5.3,<1",
        "pytest-watch>=4.1.0,<5",
        "wheel",
        "twine",
        "ipython",
    ],
    'snappy': [
        "python-snappy>=0.5.3,<1"
    ]
}


extras_require['dev'].extend(extras_require['test'])
extras_require['dev'].extend(extras_require['lint'])
extras_require['dev'].extend(extras_require['doc'])

with open('./README.md') as readme:
    long_description = readme.read()

setup(
    name='lahja',
    # *IMPORTANT*: Don't manually change the version here. Use `make bump`, as described in readme
    version='0.14.5',
    description="Generic event bus for asynchronous cross-process communication",
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='The Lahja developers',
    author_email='christoph.burgdorf@gmail.com',
    url='https://github.com/ethereum/lahja',
    include_package_data=True,
    install_requires=[
        "async-generator>=1.10,<2",
        "trio>=0.11,<0.12",
        "trio_typing>=0.2.0,<0.3.0",
    ],
    python_requires='>=3.6, <4',
    extras_require=extras_require,
    py_modules=['lahja'],
    license="MIT",
    zip_safe=False,
    keywords='eventbus asyncio trio',
    packages=find_packages(exclude=["tests", "tests.*"]),
    package_data={'lahja': ['py.typed']},
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
)



import logging
import sys


def setup_stderr_lahja_logging() -> None:
    logger = logging.getLogger("lahja")
    handler = logging.StreamHandler(sys.stderr)
    handler.setLevel(logging.DEBUG)
    logger.addHandler(handler)
    logger.setLevel(logging.DEBUG)
